import os
import shutil
from typing import *
import numpy as np
import objectbox
import time
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.vectorstores import VectorStore
from objectbox.model.entity import Entity
from objectbox.model.model import IdUid
from objectbox.model.properties import (
    VectorDistanceType,
    HnswIndex,
    Id,
    Property,
    PropertyType,
    String,
    Float32Vector
)

DIRECTORY = "objectbox"


class ObjectBox(VectorStore):
    """
    ObjectBox as Vector Store.
    To use, you should have the `objectbox` python package installed and `flatbuffers`.
    Args:
        embedding_function: Embedding function to use.
        embedding_dimensions: Dimensions of the embeddings.
        db_directory: Path to the database where data will be stored.
        clear_db: Flag for deleting all the data in the database.
        entity_model: Creates an objectbox entity.
        db: Registers the model with objectbox.
        vector_box: Initializing objectbox.
    """

    def __init__(
        self,
        embedding: Embeddings,
        embedding_dimensions: int,
        distance_type: VectorDistanceType = VectorDistanceType.EUCLIDEAN,
        db_directory: Optional[str] = None,
        clear_db: Optional[bool] = False,
        do_log: Optional[bool] = False,
    ):
        self._embedding = embedding
        self._embedding_dimensions = embedding_dimensions
        self._distance_type = distance_type
        self.db_directory = db_directory
        self._clear_db = clear_db
        self._entity_model = self._create_entity_class()
        self._db = self._create_objectbox_db()
        self._vector_box = self._db.box(self._entity_model)
        self._do_log = do_log

    @property
    def embeddings(self) -> Optional[Embeddings]:
        return self._embedding if isinstance(self._embedding, Embeddings) else None

    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        **kwargs: Any,
    ) -> List[str]:
        """Add list of text along with embeddings to the vector store
        Args:
            texts (Iterable[str]): collection of text to add to the database
        Returns:
            List of ids for the newly inserted documents
        """

        texts_list = list(texts)
        start = time.perf_counter()
        embeddings = self._embedding.embed_documents(texts_list)

        if self._do_log:
            end = time.perf_counter()
            print(f"Embedded  {len(texts_list)} documents in {end - start} seconds")
        
        ids = []

        if not metadatas:
            metadatas = [{} for _ in texts]

        start = time.perf_counter()

        with self._db.write_tx():
            for text, metadata, embedding in zip(texts, metadatas, embeddings):
                object_id = self._vector_box.put(
                    self._entity_model(
                        text=text,
                        metadata=metadata,
                        embeddings=embedding
                    )
                )
                ids.append(object_id)

        if self._do_log:
            end = time.perf_counter()
            print(f"ObjectBox stored {len(ids)} documents in { end - start} seconds")

        return ids

    def similarity_search(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> List[Document]:
        """Run similarity search on query
        Args:
            query (str): Query
            k (int): Number of results to return. Defaults to 4.
        Returns:
            List of Documents most similar to the query
        """
        embedded_query = self._embedding.embed_query(query)
        return self.similarity_search_by_vector(embedded_query, k, **kwargs)

    # Overwrite from VectorStore
    def _select_relevance_score_fn(self) -> Callable[[float], float]:
        return lambda score: self._convert_score(self._distance_type, score)

    @staticmethod
    def _convert_score(type: VectorDistanceType, score: float) -> float:
        # Map ObjectBox distance to LangChain range, in which 0 is dissimilar, 1 is most similar.
        if type == VectorDistanceType.EUCLIDEAN:
            # Not required: score = sqrt(score)  # ObjectBox returns squared Euclidean
            if score > 1.0:  # For now, we assume normalized vectors, which result in scores in the range 0..1
                return 0.0
            return 1.0 - score
        elif type == VectorDistanceType.COSINE:
            return 1.0 - score / 2.0
        elif type == VectorDistanceType.DOT_PRODUCT:
            if score > 2.0:  # For now, we assume normalized vectors, which result in scores in the range 0..2
                return 0.0
            return 1.0 - score / 2.0
        elif type == VectorDistanceType.DOT_PRODUCT_NON_NORMALIZED:
            return 1.0 - score / 2.0
        else:
            raise Exception(f"Unknown distance type: {type.name}")


    def similarity_search_with_score(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> List[Tuple[Document, float]]:
        """Run similarity search with distance.

        Args:
            query (str): Query
            k (int): Number of Documents to return. Defaults to 4.

        Returns:
            List of Documents with score most similar to the query vector.
        """
        embedded_query = self._embedding.embed_query(query)

        start = time.perf_counter()
        qb = self._vector_box.query( self._entity_model.embeddings.nearest_neighbor(embedded_query, k) )
        query = qb.build()
        results = query.find_with_scores()

        if self._do_log:
            end = time.perf_counter()
            print(f"ObjectBox retrieved {len(results)} vectors in { end - start} seconds")

        return [
            (Document(page_content=obj.text, metadata=obj.metadata), score)
            for obj, score in results
        ]

    def similarity_search_by_vector(
        self, embedding: List[float], k: int = 4, **kwargs: Any
    ) -> List[Document]:
        """Return docs most similar to embedding vector.

        Args:
            embedding (List[float]): Embedding to look up documents similar to.
            k (int): Number of Documents to return. Defaults to 4.

        Returns:
            List of Documents most similar to the query vector.
        """
        start = time.perf_counter()

        qb = self._vector_box.query( self._entity_model.embeddings.nearest_neighbor(embedding, k) )
        query = qb.build()
        results = query.find_with_scores()

        if self._do_log:
            end = time.perf_counter()
            print(f"ObjectBox retrieved {len(results)} vectors in { end - start} seconds")

        return [
            Document(page_content=obj.text, metadata=obj.metadata)
            for obj, _ in results
        ]

    @classmethod
    def from_texts(
        cls,
        texts: List[str],
        embedding: Embeddings,
        metadatas: Optional[List[dict]] = None,
        **kwargs: Any,
    ) -> objectbox:
        """Create ObjectBox from list of text
        Args:
            texts (List[str]): list of text to vectorize and store
            embedding (Optional[Embeddings]): Embedding function.
        Returns:
            ObjectBox object initialized and ready for use."""
        ob = cls(embedding, **kwargs)
        ob.add_texts(texts, metadatas)
        return ob

    def delete(self, ids: Optional[List[str]] = None, **kwargs: Any) -> bool:
        """Delete by vector ID.

        Args:
            ids: List of ids to delete.

        Returns:
            bool: True if deletion is successful,
            False otherwise, None if not implemented.
        """
        with self._db.write_tx():
            for id in ids:
                try:
                    int_id = int(id)
                except ValueError:
                    raise ValueError("invalid id input")
                self._vector_box.remove(int_id)

        if self._do_log:
            end = time.perf_counter()
            print(f"ObjectBox deleted {len(ids)} vectors in { end - start} seconds")

        return True

    def _create_objectbox_db(self) -> objectbox:
        """registering the VectorEntity model and setting up objectbox database"""
        db_path = DIRECTORY if self.db_directory is None else self.db_directory
        if self._clear_db and os.path.exists(db_path):
            shutil.rmtree(db_path)
        model = objectbox.Model()
        model.entity(self._entity_model)
        return objectbox.Store(model=model,directory=db_path)

    def _create_entity_class(self) -> Entity:
        """Dynamically define an Entity class according to the parameters."""

        @Entity()
        class VectorEntity:
            id = Id()
            text = String()
            metadata = Property(dict, type=PropertyType.flex, id=3, uid=1003)
            embeddings = Float32Vector(index=HnswIndex(dimensions=self._embedding_dimensions, distance_type=self._distance_type))

        return VectorEntity
